@@ -1,3 +1,29 @@
+"""
+This module provides functionality for parsing URLs for Dandisets & assets on
+Dandi Archive servers and for fetching the objects to which the URLs refer.
+See the docstring for `parse_dandi_url()` for a list of accepted URL formats.
+
+Basic operation begins by calling `parse_dandi_url()` on a URL in order to
+acquire a `ParsedDandiURL` instance, which can then be used to obtain the
+Dandiset and/or assets specified in the URL.  Call an instance's
+`~ParsedDandiURL.get_dandiset()` and/or `~ParsedDandiURL.get_assets()` to get
+the assets, passing in a `~dandi.dandiapi.DandiAPIClient` for the appropriate
+Dandi Archive API instance; an unauthenticated client pointing to the correct
+instance can be acquired via the `~ParsedDandiURL.get_client()` method.  As a
+convenience, one can acquire a client, the Dandiset, and an iterator of all
+assets by using the `~ParsedDandiAPI.navigate()` context manager like so:
+
+.. code:: python
+
+    parsed_url = parse_dandi_url("https://...")
+    with parsed_url.navigate() as (client, dandiset, assets):
+        ...
+    # The client's session is closed when the context manager exits.
+
+As a further convenience, a URL can be parsed and navigated in one fell swoop
+using the `navigate_url()` function.
+"""
+
 from abc import ABC, abstractmethod
 from contextlib import contextmanager
 import re
@@ -9,7 +35,13 @@ from pydantic import AnyHttpUrl, BaseModel, parse_obj_as, validator
 import requests
 
 from . import get_logger
-from .consts import RETRY_STATUSES, VERSION_REGEX, known_instances
+from .consts import (
+    DANDISET_ID_REGEX,
+    PUBLISHED_VERSION_REGEX,
+    RETRY_STATUSES,
+    VERSION_REGEX,
+    known_instances,
+)
 from .dandiapi import BaseRemoteAsset, DandiAPIClient, RemoteDandiset
 from .exceptions import FailedToConnectError, NotFoundError, UnknownURLError
 from .utils import get_instance
@@ -23,18 +55,18 @@ class ParsedDandiURL(ABC, BaseModel):
     (Dandiset or asset(s)).  Subclasses must implement `get_assets()`.
 
     Most methods take a ``client: DandiAPIClient`` argument, which must be a
-    `DandiAPIClient` object for querying `api_url` (This is not checked).  Such
-    a client instance can be obtained by calling `get_client()`, or an
-    appropriate pre-existing client instance can be passed instead.
-    """
+    `~dandi.dandiapi.DandiAPIClient` object for querying `api_url` (This is not
+    checked).  Such a client instance can be obtained by calling
+    `get_client()`, or an appropriate pre-existing client instance can be
+    passed instead."""
 
     #: The base URL of the Dandi API service, without a trailing slash
     api_url: AnyHttpUrl
     #: The ID of the Dandiset given in the URL
     dandiset_id: Optional[str]
-    #: The version of the Dandiset, if specified.  If this is not set, methods
-    #: that need the Dandiset version will call `get_version_id()` to get an
-    #: appropriate default value.
+    #: The version of the Dandiset, if specified.  If this is not set, the
+    #: version will be defaulted using the rules described under
+    #: `DandiAPIClient.get_dandiset()`.
     version_id: Optional[str] = None
 
     @validator("api_url")
@@ -42,44 +74,47 @@ class ParsedDandiURL(ABC, BaseModel):
         return parse_obj_as(AnyHttpUrl, v.rstrip("/"))
 
     def get_client(self) -> DandiAPIClient:
-        """Returns an unauthenticated `DandiAPIClient` for `api_url`"""
+        """
+        Returns an unauthenticated `~dandi.dandiapi.DandiAPIClient` for
+        `api_url`
+        """
         return DandiAPIClient(self.api_url)
 
-    def get_dandiset(self, client: DandiAPIClient) -> Optional[RemoteDandiset]:
+    def get_dandiset(
+        self, client: DandiAPIClient, lazy: bool = True
+    ) -> Optional[RemoteDandiset]:
         """
         Returns information about the specified (or default) version of the
-        specified Dandiset
+        specified Dandiset.  Returns `None` if the URL did not contain a
+        Dandiset identifier.
+
+        If ``lazy`` is true, a "lazy" `RemoteDandiset` instance is returned,
+        with no requests made until any data is actually required.
         """
         if self.dandiset_id is not None:
-            return client.get_dandiset(self.dandiset_id, self.get_version_id(client))
+            return client.get_dandiset(self.dandiset_id, self.version_id, lazy=lazy)
         else:
             return None
 
-    def get_version_id(self, client: DandiAPIClient) -> str:
-        """
-        Returns `version_id` or determines a default value if unset.
-
-        If `version_id` is non-`None`, returns it.  Otherwise, the ID of the
-        most recent published version of the Dandiset is returned, if any,
-        otherwise the ID of the draft version is returned.
-        """
-        if self.dandiset_id is None:
-            raise ValueError("Cannot get version for Dandiset-less URL")
-        elif self.version_id is None:
-            r = client.get(f"/dandisets/{self.dandiset_id}/")
-            version = r["most_recent_published_version"] or r["draft_version"]
-            return version["version"]
-        else:
-            return self.version_id
-
     @abstractmethod
-    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
+    def get_assets(
+        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
+    ) -> Iterator[BaseRemoteAsset]:
         """
         Returns an iterator of asset structures for the assets referred to by
         or associated with the URL.  For a URL pointing to just a Dandiset,
         this is the set of all assets in the given or default version of the
         Dandiset.  For a URL that specifies a specific asset or collection of
         assets in a Dandiset, this is all of those assets.
+
+        When multiple assets are returned, they can be sorted by a given field
+        by passing the name of that field as the ``order`` parameter.  The
+        accepted field names are ``"created"``, ``"modified"``, and ``"path"``.
+        Prepend a hyphen to the field name to reverse the sort order.
+
+        If ``strict`` is true, then fetching assets for a URL that refers to a
+        nonexistent resource will raise a `NotFoundError`; if it is false, the
+        method will instead return an empty iterator.
         """
         ...
 
@@ -93,20 +128,28 @@ class ParsedDandiURL(ABC, BaseModel):
 
     @contextmanager
     def navigate(
-        self,
+        self, strict: bool = False
     ) -> Iterator[
         Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterator[BaseRemoteAsset]]
     ]:
         """
-        A context manager that returns a triple of a `DandiAPIClient` (with an
-        open session that is closed when the context manager closes), the
-        return value of `get_dandiset()`, and the return value of
-        `get_assets()`
+        A context manager that returns a triple of a
+        `~dandi.dandiapi.DandiAPIClient` (with an open session that is closed
+        when the context manager closes), the return value of `get_dandiset()`,
+        and the return value of `get_assets()`.
+
+        If ``strict`` is true, then `get_dandiset()` is called with
+        ``lazy=False`` and `get_assets()` is called with ``strict=True``; if
+        ``strict`` is false, the opposite occurs.
         """
         # We could later try to "dandi_authenticate" if run into permission
         # issues.  May be it could be not just boolean but the "id" to be used?
         with self.get_client() as client:
-            yield (client, self.get_dandiset(client), self.get_assets(client))
+            yield (
+                client,
+                self.get_dandiset(client, lazy=not strict),
+                self.get_assets(client, strict=strict),
+            )
 
 
 class DandisetURL(ParsedDandiURL):
@@ -114,9 +157,14 @@ class DandisetURL(ParsedDandiURL):
     Parsed from a URL that only refers to a Dandiset (possibly with a version)
     """
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
+    def get_assets(
+        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
+    ) -> Iterator[BaseRemoteAsset]:
         """Returns all assets in the Dandiset"""
-        return self.get_dandiset(client).get_assets()
+        with _maybe_strict(strict):
+            yield from self.get_dandiset(client, lazy=not strict).get_assets(
+                order=order
+            )
 
 
 class SingleAssetURL(ParsedDandiURL):
@@ -139,18 +187,16 @@ class BaseAssetIDURL(SingleAssetURL):
 
     asset_id: str
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
+    def get_assets(
+        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
+    ) -> Iterator[BaseRemoteAsset]:
         """
-        Yields the asset with the given ID.  Yields nothing if the asset does
-        not exist.
+        Yields the asset with the given ID.  If the asset does not exist, then
+        a `NotFoundError` is raised if ``strict`` is true, and nothing is
+        yielded if ``strict`` is false.
         """
-        try:
+        with _maybe_strict(strict):
             yield client.get_asset(self.asset_id)
-        except requests.HTTPError as e:
-            if e.response.status_code == 404:
-                return
-            else:
-                raise
 
     def get_asset_ids(self, client: DandiAPIClient) -> Iterator[str]:
         """Yields the ID of the asset (regardless of whether it exists)"""
@@ -164,18 +210,16 @@ class AssetIDURL(SingleAssetURL):
 
     asset_id: str
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
+    def get_assets(
+        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
+    ) -> Iterator[BaseRemoteAsset]:
         """
-        Yields the asset with the given ID.  Yields nothing if the asset does
-        not exist.
+        Yields the asset with the given ID.  If the Dandiset or asset does not
+        exist, then a `NotFoundError` is raised if ``strict`` is true, and
+        nothing is yielded if ``strict`` is false.
         """
-        try:
-            yield self.get_dandiset(client).get_asset(self.asset_id)
-        except requests.HTTPError as e:
-            if e.response.status_code == 404:
-                return
-            else:
-                raise
+        with _maybe_strict(strict):
+            yield self.get_dandiset(client, lazy=not strict).get_asset(self.asset_id)
 
     def get_asset_ids(self, client: DandiAPIClient) -> Iterator[str]:
         """Yields the ID of the asset (regardless of whether it exists)"""
@@ -187,9 +231,14 @@ class AssetPathPrefixURL(MultiAssetURL):
     Parsed from a URL that refers to a collection of assets by path prefix
     """
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
+    def get_assets(
+        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
+    ) -> Iterator[BaseRemoteAsset]:
         """Returns the assets whose paths start with `path`"""
-        return self.get_dandiset(client).get_assets_with_path_prefix(self.path)
+        with _maybe_strict(strict):
+            yield from self.get_dandiset(
+                client, lazy=not strict
+            ).get_assets_with_path_prefix(self.path, order=order)
 
 
 class AssetItemURL(SingleAssetURL):
@@ -197,27 +246,46 @@ class AssetItemURL(SingleAssetURL):
 
     path: str
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
+    def get_assets(
+        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
+    ) -> Iterator[BaseRemoteAsset]:
         """
-        Yields the asset whose path equals `path`.  If there is no such asset,
-        this method yields nothing, unless the path happens to be the path to
-        an asset directory, in which case an error is raised indicating that
-        the user left off a trailing slash.
+        Yields the asset whose path equals `path`.  If there is no such asset:
+
+        - If ``strict`` is true, a `NotFoundError` is raised.
+        - If ``strict`` is false and the path happens to be the path to an
+          asset directory, a `ValueError` is raised indicating that the user
+          left off a trailing slash.
+        - Otherwise, nothing is yielded.
         """
-        d = self.get_dandiset(client)
         try:
-            asset = d.get_asset_by_path(self.path)
+            dandiset = self.get_dandiset(client, lazy=not strict)
+            # Force evaluation of the version here instead of when
+            # get_asset_by_path() is called so we don't get nonexistent
+            # dandisets with unspecified versions mixed up with nonexistent
+            # assets.
+            dandiset.version_id
         except NotFoundError:
+            if strict:
+                raise
+            else:
+                return
+        try:
+            yield dandiset.get_asset_by_path(self.path)
+        except NotFoundError:
+            if strict:
+                raise
             try:
-                next(d.get_assets_with_path_prefix(self.path + "/"))
+                next(dandiset.get_assets_with_path_prefix(self.path + "/"))
+            except NotFoundError:
+                # Dandiset has explicit version that doesn't exist.
+                return
             except StopIteration:
                 pass
             else:
                 raise ValueError(
                     f"Asset path {self.path!r} points to a directory but lacks trailing /"
                 )
-        else:
-            yield asset
 
 
 class AssetFolderURL(MultiAssetURL):
@@ -227,7 +295,9 @@ class AssetFolderURL(MultiAssetURL):
 
     path: str
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
+    def get_assets(
+        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
+    ) -> Iterator[BaseRemoteAsset]:
         """
         Returns all assets under the folder at `path`.  Yields nothing if the
         folder does not exist.
@@ -235,21 +305,45 @@ class AssetFolderURL(MultiAssetURL):
         path = self.path
         if not path.endswith("/"):
             path += "/"
-        return self.get_dandiset(client).get_assets_with_path_prefix(path)
+        with _maybe_strict(strict):
+            yield from self.get_dandiset(
+                client, lazy=not strict
+            ).get_assets_with_path_prefix(path, order=order)
 
 
 @contextmanager
-def navigate_url(url):
-    """Context manager to 'navigate' URL pointing to DANDI archive.
+def _maybe_strict(strict: bool) -> Iterator[None]:
+    try:
+        yield
+    except NotFoundError:
+        if strict:
+            raise
 
-    :param str url: URL which might point to a dandiset, a folder, or an
-        asset(s)
 
-    :returns: Generator of one ``(client, dandiset, assets)``; ``client`` will
-        have established a session for the duration of the context
+@contextmanager
+def navigate_url(
+    url: str, strict: bool = False
+) -> Iterator[
+    Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterator[BaseRemoteAsset]]
+]:
+    """
+    A context manager that takes a URL pointing to a DANDI Archive and
+    returns a triple of a `~dandi.dandiapi.DandiAPIClient` (with an open
+    session that is closed when the context manager closes), the Dandiset
+    identified in the URL (if any), and the assets specified by the URL (or, if
+    no specific assets were specified, all assets in the Dandiset).
+
+    :param str url: URL which might point to a Dandiset, folder, or asset(s)
+    :param bool strict:
+        If true, then `get_dandiset()` is called with ``lazy=False`` and
+        `get_assets()` is called with ``strict=True``; if false, the opposite
+        occurs.
+    :returns: Context manager that yields a ``(client, dandiset, assets)``
+        tuple; ``client`` will have a session established for the duration of
+        the context
     """
     parsed_url = parse_dandi_url(url)
-    with parsed_url.navigate() as (client, dandiset, assets):
+    with parsed_url.navigate(strict=strict) as (client, dandiset, assets):
         yield (client, dandiset, assets)
 
 
@@ -257,7 +351,7 @@ class _dandi_url_parser:
     # Defining as a class with all the attributes to not leak all the variables
     # etc into module space, and later we might end up with classes for those
     # anyways
-    dandiset_id_grp = "(?P<dandiset_id>[0-9]{6})"
+    dandiset_id_grp = f"(?P<dandiset_id>{DANDISET_ID_REGEX})"
     # Should absorb port and "api/":
     server_grp = "(?P<server>(?P<protocol>https?)://(?P<hostname>[^/]+)/(api/)?)"
     known_urls = [
@@ -275,9 +369,12 @@ class _dandi_url_parser:
         # TODO: Later should better conform to our API, so we could allow
         #       for not only "dandiarchive.org" URLs
         (
-            re.compile(r"DANDI:.*"),
-            {"rewrite": lambda x: "https://identifiers.org/" + x},
-            "DANDI:<dandiset id>",
+            re.compile(
+                fr"DANDI:{DANDISET_ID_REGEX}(?:/{PUBLISHED_VERSION_REGEX})?",
+                flags=re.I,
+            ),
+            {"rewrite": lambda x: "https://identifiers.org/" + x.lower()},
+            "DANDI:<dandiset id>[/<version id>]",
         ),
         (
             re.compile(r"https?://dandiarchive\.org/.*"),
@@ -285,9 +382,13 @@ class _dandi_url_parser:
             "https://dandiarchive.org/...",
         ),
         (
-            re.compile(r"https?://identifiers\.org/DANDI:.*"),
+            re.compile(
+                fr"https?://identifiers\.org/DANDI:{DANDISET_ID_REGEX}"
+                fr"(?:/{PUBLISHED_VERSION_REGEX})?",
+                flags=re.I,
+            ),
             {"handle_redirect": "pass"},
-            "https://identifiers.org/DANDI:<dandiset id>",
+            "https://identifiers.org/DANDI:<dandiset id>[/<version id>]",
         ),
         (
             re.compile(
@@ -310,7 +411,7 @@ class _dandi_url_parser:
         (
             re.compile(
                 rf"{server_grp}(?P<asset_type>dandiset)s/{dandiset_id_grp}"
-                rf"(/(versions(/(?P<version>{VERSION_REGEX}))?)?)?"
+                rf"(/(versions(/(?P<version>{VERSION_REGEX}))?/?)?)?"
             ),
             {},
             "https://<server>[/api]/dandisets/<dandiset id>[/versions[/<version>]]",
@@ -349,7 +450,7 @@ class _dandi_url_parser:
         # use
         (
             re.compile(
-                rf"dandi://(?P<instance_name>({'|'.join(known_instances)}))"
+                rf"dandi://(?P<instance_name>[-\w._]+)"
                 rf"/{dandiset_id_grp}"
                 rf"(@(?P<version>{VERSION_REGEX}))?"
                 rf"(/(?P<location>.*)?)?"
@@ -365,40 +466,74 @@ class _dandi_url_parser:
             "https://<server>/...",
         ),
     ]
-    known_patterns = "Patterns for known setups:" + "\n - ".join(
+    known_patterns = "Accepted resource identifier patterns:" + "\n - ".join(
         [""] + [display for _, _, display in known_urls]
     )
     map_to = {}
     for (gui, redirector, api) in known_instances.values():
-        for h in (gui, redirector):
-            if h and api:
-                map_to[h] = api
+        if api:
+            for h in (gui, redirector):
+                if h:
+                    map_to[h] = api
 
     @classmethod
-    def parse(cls, url, *, map_instance=True):
+    def parse(cls, url: str, *, map_instance: bool = True) -> ParsedDandiURL:
         """
-        Parse url like and return server (address), asset_id and/or directory
+        Parse a Dandi Archive URL and return a `ParsedDandiURL` instance.
 
-        Example URLs (as of 20210428):
+        The accepted URL formats are as follows.  Text in [brackets] is
+        optional.  A ``server`` field is a base API, GUI, or redirector URL for
+        a registered Dandi Archive instance.  If an optional ``version`` field
+        is omitted from a URL, the given Dandiset's most recent published
+        version, if any, or else its draft version will be used.
 
-        - Dataset landing page metadata:
-          https://gui.dandiarchive.org/#/dandiset/000003
+        - :samp:`https://identifiers.org/DANDI:{dandiset-id}[/{version}]`
+          (case insensitive; ``version`` cannot be "draft") when it redirects
+          to one of the other URL formats
 
-        Individual and multiple files:
+        - :samp:`DANDI:{dandiset-id}[/{version}]` — (case insensitive;
+          ``version`` cannot be "draft") Abbreviation for the above
 
-        - dandi???
+        - Any ``https://dandiarchive.org/`` or
+          ``https://*dandiarchive-org.netflify.app/`` URL which redirects to
+          one of the other URL formats
 
-        Multiple selected files + folders -- we do not support ATM, then further
-        RFing would be due, probably making this into a generator or returning a
-        list of entries.
+        - :samp:`https://{server}[/api]/[#/]dandiset/{dandiset-id}[/{version}][/files[?location={path}]]`
+          — If ``path`` is not specified, the URL refers to a Dandiset and is
+          converted to a `DandisetURL`.  If ``path`` is specified and it ends
+          with a forward slash, the URL refers to an asset folder and is
+          converted to an `AssetFolderURL` instance; if it does not end with a
+          slash, it refers to a single asset and is converted to an
+          `AssetItemURL` instance.
 
-        "Features":
+        - :samp:`https://{server}[/api]/dandisets/{dandiset-id}[/versions[/{version}]]`
+          — Refers to a Dandiset and is converted to a `DandisetURL`
 
-        - uses some of `known_instances` to map some urls, e.g. from
-          gui.dandiarchive.org ones into girder.
+        - :samp:`https://{server}[/api]/assets/{asset-id}[/download]` — Refers
+          to a single asset by identifier and is converted to a
+          `BaseAssetIDURL`
 
-        :rtype: ParsedDandiURL
+        - :samp:`https://{server}[/api]/dandisets/{dandiset-id}/versions/{version}/assets/{asset-id}[/download]`
+          — Refers to a single asset and is converted to an `AssetIDURL`
+
+        - :samp:`https://{server}[/api]/dandisets/{dandiset-id}/versions/{version}/assets/?path={path}`
+          — Refers to all assets in the given Dandiset whose paths begin with
+          the prefix ``path``; converted to an `AssetPathPrefixURL`
+
+        - :samp:`dandi://{instance-name}/{dandiset-id}[@{version}][/{path}]`,
+          where ``instance-name`` is the name of a registered Dandi Archive
+          instance.  If ``path`` is not specified, the URL refers to a Dandiset
+          and is converted to a `DandisetURL`.  If ``path`` is specified and it
+          ends with a forward slash, the URL refers to an asset folder and is
+          converted to an `AssetFolderURL` instance; if it does not end with a
+          slash, it refers to a single asset and is converted to an
+          `AssetItemURL` instance.
+
+        - Any other HTTPS URL that redirects to one of the above
+
+        :raises UnknownURLError: if the URL is not one of the above
         """
+
         lgr.debug("Parsing url %s", url)
 
         # Loop through known url regexes and stop as soon as one is matching
@@ -444,7 +579,13 @@ class _dandi_url_parser:
                     parsed_url.api_url = known_instance.api
                 continue  # in this run we ignore and match further
             elif "instance_name" in groups:
-                known_instance = get_instance(groups["instance_name"])
+                try:
+                    known_instance = get_instance(groups["instance_name"])
+                except KeyError:
+                    raise UnknownURLError(
+                        f"Unknown instance {groups['instance_name']!r}.  Valid instances: "
+                        + ", ".join(sorted(known_instances))
+                    )
                 assert known_instance.api  # must be defined
                 groups["server"] = known_instance.api
                 # could be overloaded later depending if location is provided
@@ -475,10 +616,9 @@ class _dandi_url_parser:
         path = groups.get("path")
         if location:
             location = urlunquote(location)
-            # ATM carries leading '/' which IMHO is not needed/misguiding somewhat, so
-            # I will just strip it
+            # ATM carries leading '/' which IMHO is not needed/misguiding
+            # somewhat, so I will just strip it
             location = location.lstrip("/")
-
         # if location is not degenerate -- it would be a folder or a file
         if location:
             if location.endswith("/"):
@@ -522,7 +662,14 @@ class _dandi_url_parser:
         return parsed_url
 
     @staticmethod
-    def follow_redirect(url):
+    def follow_redirect(url: str) -> str:
+        """
+        Resolve the given URL by following all redirects.
+
+        :raises NotFoundError: if a 404 response is returned
+        :raises FailedToConnectError: if a response other than 200, 400, 404,
+            or one of the statuses in `~dandi.consts.RETRY_STATUSES` is returned
+        """
         i = 0
         while True:
             r = requests.head(url, allow_redirects=True)
@@ -540,6 +687,7 @@ class _dandi_url_parser:
                 )
             elif r.url != url:
                 return r.url
+            assert isinstance(url, str)
             return url
 
 
