@@ -15,6 +15,7 @@ from . import get_logger
 from .consts import RETRY_STATUSES, dandiset_metadata_file
 from .dandiarchive import DandisetURL, MultiAssetURL, SingleAssetURL, parse_dandi_url
 from .dandiset import Dandiset
+from .exceptions import NotFoundError
 from .support.digests import get_digest
 from .support.pyout import naturalsize
 from .utils import (
@@ -26,6 +27,7 @@ from .utils import (
     on_windows,
     path_is_subpath,
     pluralize,
+    yaml_load,
 )
 
 lgr = get_logger()
@@ -178,7 +180,7 @@ def download_generator(
 
     """
 
-    with parsed_url.navigate() as (client, dandiset, assets):
+    with parsed_url.navigate(strict=True) as (client, dandiset, assets):
         if assets_it:
             assets_it.gen = assets
             assets = assets_it
@@ -195,19 +197,6 @@ def download_generator(
             return
 
         for asset in assets:
-            metadata = asset.get_raw_metadata()
-            d = metadata.get("digest", {})
-            if "dandi:dandi-etag" in d:
-                digests = {"dandi-etag": d["dandi:dandi-etag"]}
-            else:
-                raise RuntimeError(
-                    f"dandi-etag not available for asset. Known digests: {d}"
-                )
-            try:
-                digests["sha256"] = d["dandi:sha2-256"]
-            except KeyError:
-                pass
-
             path = asset.path.lstrip("/")  # make into relative path
             path = op.normpath(path)
             if not isinstance(parsed_url, DandisetURL):
@@ -224,6 +213,23 @@ def download_generator(
                     )
             download_path = op.join(output_path, path)
 
+            try:
+                metadata = asset.get_raw_metadata()
+            except NotFoundError as e:
+                yield {"path": path, "status": "error", "message": str(e)}
+                continue
+            d = metadata.get("digest", {})
+            if "dandi:dandi-etag" in d:
+                digests = {"dandi-etag": d["dandi:dandi-etag"]}
+            else:
+                raise RuntimeError(
+                    f"dandi-etag not available for asset. Known digests: {d}"
+                )
+            try:
+                digests["sha256"] = d["dandi:sha2-256"]
+            except KeyError:
+                pass
+
             downloader = asset.get_download_file_iter()
 
             try:
@@ -384,8 +390,13 @@ def _populate_dandiset_yaml(dandiset_path, dandiset, existing):
     lgr.debug("Updating %s from obtained dandiset metadata", dandiset_metadata_file)
     mtime = dandiset.modified
     if op.lexists(dandiset_yaml):
+        with open(dandiset_yaml) as fp:
+            if yaml_load(fp, typ="safe") == metadata:
+                yield _skip_file("no change")
+                return
         if existing == "error":
-            raise FileExistsError(dandiset_yaml)
+            yield {"status": "error", "message": "already exists"}
+            return
         elif existing == "refresh" and op.lexists(
             op.join(dandiset_path, ".git", "annex")
         ):
@@ -396,10 +407,10 @@ def _populate_dandiset_yaml(dandiset_path, dandiset, existing):
         ):
             yield _skip_file("already exists")
             return
-    dandiset = Dandiset(dandiset_path, allow_empty=True)
-    dandiset.path_obj.mkdir(exist_ok=True)  # exist_ok in case of parallel race
-    old_metadata = dandiset.metadata
-    dandiset.update_metadata(metadata)
+    ds = Dandiset(dandiset_path, allow_empty=True)
+    ds.path_obj.mkdir(exist_ok=True)  # exist_ok in case of parallel race
+    old_metadata = ds.metadata
+    ds.update_metadata(metadata)
     os.utime(dandiset_yaml, (time.time(), mtime.timestamp()))
     yield {
         "status": "done",
